/******************************************************************************
 * $Source: /export/CVS/java/minissl/src/de/bb/minissl/BigInteger.java,v $
 * $Revision: 1.4 $
 * $Date: 2003/01/04 12:22:35 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 * 
 * Copyright (c) by Stefan Bebbo Franke 1999-2000.
 * All rights reserved
 *
 * fast and small BigInteger implementation
 *
 ******************************************************************************
 NON COMMERCIAL PUBLIC LICENSE
 ******************************************************************************

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Every product and solution using this software, must be free
 of any charge. If the software is used by a client part, the
 server part must also be free and vice versa.

 2. Each redistribution must retain the copyright notice, and
 this list of conditions and the following disclaimer.

 3. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with the
 distribution.

 4. All advertising materials mentioning features or use of this
 software must display the following acknowledgment:
 "This product includes software developed by BebboSoft,
 written by Stefan Bebbo Franke. (http://www.bebbosoft.de)"

 5. Redistributions of any form whatsoever must retain the following
 acknowledgment:
 "This product includes software developed by BebboSoft,
 written by Stefan Bebbo Franke. (http://www.bebbosoft.de)"

 ******************************************************************************
 DISCLAIMER OF WARRANTY

 Software is provided "AS IS," without a warranty of any kind.
 You may use it on your own risk.

 ******************************************************************************
 LIMITATION OF LIABILITY

 I SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY YOU OR ANY THIRD PARTY
 AS A RESULT OF USING OR DISTRIBUTING SOFTWARE. IN NO EVENT WILL I BE LIABLE
 FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS
 OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE
 SOFTWARE, EVEN IF I HAVE ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

 *****************************************************************************
 COPYRIGHT

 (c) 1994-2000 by BebboSoft, Stefan "Bebbo" Franke, all rights reserved

 *****************************************************************************/

package de.bb.minissl;

import java.util.*;

/**
 * A small and fast implementation for BigInteger.
 * Like the String class, all  functions do not modify an existing BigInteger,
 * but return a new created BigInteger with the result.
 */
public class IntImpl //extends java.lang.Number
{
  final static IntImpl ZERO = new IntImpl();

  final static IntImpl ONE = new IntImpl(1);

  final static IntImpl TWO = new IntImpl(2);

  final static IntImpl THREE = new IntImpl(3);

  final static IntImpl TEN = new IntImpl(10);

  private int[] data; // data for BigInteger

  private int len; // used range data

  private boolean positive; // sign of current number

  private static IntImpl pCheck[]; // a couple of small primes
  /** 
   * Init the small primes.
   */
  static {
    
    pCheck = new IntImpl[40];
    IntImpl p = IntImpl.ONE;
    IntImpl q[];
    for (int i = 0; i < pCheck.length;) {
      p = p.add(TWO);
      for (IntImpl t = THREE;; t = t.add(TWO)) {
        q = p.divideAndRemainder(t);
        if (q[0].compareTo(t) < 0) {
          pCheck[i++] = p;
          break;
        }
        if (q[1].compareTo(IntImpl.ZERO) == 0)
          break;
      }
    }
    
  }

  /**
   * default ct, constructs a BigInteger with value 0.
   */
  public IntImpl()
  {
    this(0);
  }

  /**
   * Create a prime number for given bit length.
   * @param bitLength the desired bit length of the prime.
   * @param certainty how sure is it a prim P = 1 - (1/2)^certainty.
   * @param rand the random generator for random numbers.
   */
  public IntImpl(int bitLength, int certainty, Random rand)
  {
    for (;;) {
      makeRandom(bitLength, rand);
      data[0] |= 1;
      int i = 0;
      for (; i < pCheck.length; ++i)
        if (this.mod(pCheck[i]).len == 0)
          break;
      if (i < pCheck.length)
        continue;
      if (this.isProbablePrime(rand, certainty))
        break;
    }
  }

  /**
   * Create a random number for given bit length.
   * @param bitLength the desired bit length of the number.
   * @param rand the random generator for random numbers.
   */
  IntImpl(int bitLength, Random rand)
  {
    makeRandom(bitLength, rand);
  }

  /**
   * Internal function which allocates a new int array and fills it with random bits.
   * @param bitLength the desired bit length of the number.
   * @param rand the random generator for random numbers.
   */
  private void makeRandom(int bitLength, Random rand)
  {
    data = new int[(bitLength + 31) >>> 5];
    len = data.length;
    if (bitLength == 0)
      return;
    for (int i = 0; i < len; ++i) {
      data[i] = rand.nextInt();
    }
    bitLength &= 31;
    if (bitLength == 0)
      bitLength = 32;
    int mask = (int) ((1L << bitLength) - 1);
    --bitLength;
    data[len - 1] = (data[len - 1] & mask) | (1 << bitLength);
    positive = true;
  }

  /**
   * Create a BigInteger from a long value.
   * @param l the long value which is assigned to this BigInteger.
   */
  IntImpl(long l)
  {
    data = new int[2];
    positive = l >= 0;
    if (!positive)
      l = -l;

    data[0] = (int) l;
    data[1] = (int) (l >>> 32);
    len = 2;
    fix00();
  }

  /**
   * Create a BigInteger from a String, the String is expected to contain some number with radix 10.
   * @param s input string, containing a number with radix 10.
   */
  IntImpl(String s)
  {
    this(s, 10);
  }

  /**
   * Create a BigInteger from a String, the String is expected to contain some number with the given radix.
   * @param s input string, containing a number with the given radix.
   * @param radix the radix to use, right now 10 and 16 are supported.
   */
  public IntImpl(String s, int radix)
  {
    this(radix == 10 ? toBytes(s) : toBytesfromHex(s));
  }

  /**
   * Create a BigInteger from a byte array, which is interpreted as a big endian number.
   * If the highest bit is set, the number is treated as a negative number.
   * @param d input byte array.
   */
  public IntImpl(byte d[])
  {
    this(d, d.length == 0 ? true : (d[0] & 0x80) == 0);
  }
  IntImpl(byte d[], boolean pos)
  {
    positive = pos;
    data = new int[1 + (d.length + 3) / 4];
    for (int i = d.length, j = 0; i > 0;) {
      int k = (0xff & d[--i]);
      if (i > 0) {
        k |= (0xff & d[--i]) << 8;
        if (i > 0) {
          k |= (0xff & d[--i]) << 16;
          if (i > 0) {
            k |= (0xff & d[--i]) << 24;
          }
        }
      }
      data[j++] = k;
    }
    len = data.length;
    fix00();
  }

  /**
   * Internal ct to create a BigInteger from an int array, its used length and a sign flag.
   * @param odata the data which is used by that BigInteger.
   * @param olen used length in the given array.
   * @param opos true, when the number is positive.
   */
  IntImpl(int odata[], int olen, boolean opos)
  {
    data = odata;
    len = olen;
    positive = opos;
    fix00();
  }

  /**
   * Convert this BigInteger into an hex string.
   * @return a new String containing the hex value of this BigInteger.
   */
  public String toHexString()
  {
    StringBuffer r = new StringBuffer();
    if (!positive)
      r.append('-');
    for (int i = len - 1; i >= 0; --i) {
      String n = Integer.toHexString(data[i]);
      if (i < len - 1)
      while (n.length() < 8)
        n = "0" + n;
      r.append(n);
/*      
      int z = data[i] >>> 24;
      if (z != 0 || !first) {
        first = false;
        int h = (z >>> 4) & 0xf;
        r = r.append((char) (h > 9 ? h + 'A' - 10 : h + '0'));
        z &= 0xf;
        r = r.append((char) (z > 9 ? z + 'A' - 10 : z + '0'));
      }
      z = data[i] >>> 16;
      if (z != 0 || !first) {
        first = false;
        int h = (z >>> 4) & 0xf;
        r = r.append((char) (h > 9 ? h + 'A' - 10 : h + '0'));
        z &= 0xf;
        r = r.append((char) (z > 9 ? z + 'A' - 10 : z + '0'));
      }
      z = data[i] >>> 8;
      if (z != 0 || !first) {
        first = false;
        int h = (z >>> 4) & 0xf;
        r = r.append((char) (h > 9 ? h + 'A' - 10 : h + '0'));
        z &= 0xf;
        r = r.append((char) (z > 9 ? z + 'A' - 10 : z + '0'));
      }
      z = data[i];
      if (z != 0 || !first) {
        first = false;
        int h = (z >>> 4) & 0xf;
        r = r.append((char) (h > 9 ? h + 'A' - 10 : h + '0'));
        z &= 0xf;
        r = r.append((char) (z > 9 ? z + 'A' - 10 : z + '0'));
      }
*/      
    }
    return r.length() == 0 ? "00" : new String(r);
  }

  /**
   * Convert this BigInteger into a decimal string.
   * @return a new String containing the decimal value of this BigInteger.
   */
  public String toString()
  {
    StringBuffer r = new StringBuffer();
    IntImpl z = this;
    if (z.bitLength() == 0) r.append("0");
    while (z.bitLength() != 0) {
      IntImpl d = z.mod(TEN);
      z = z.divide(TEN);
      r.append((char) (d.data[0] + '0'));
    }

    if (!positive)
      r.append('-');
    return new String(r.reverse());
  }

  /**
   * Internal function to suppress leading zeros.
   */
  private void fix00()
  {
    for (; len > 0; --len) {
      if (data[len - 1] != 0)
        break;
    }
  }

  /**
   * Internal function to suppress leading ff on negative numbers.
   */
  private void fixFF()
  {
    for (int i = len - 1; i >= 0; --i)
      data[i] = -data[i];
    fix00();
  }

  /**
   * Clone this BigInteger.
   * @return a new BigInteger containing the same value.
   */
  public Object clone()
  {
    int[] odata = new int[len];
    System.arraycopy(data, 0, odata, 0, len);
    return new IntImpl(odata, len, positive);
  }

  /**
   * Convert this BigInteger into a byte array.
   * @return a new alloctaed byte array containing the bytes of this BigInteger.
   */
  public byte[] toByteArray()
  {
    if (len == 0)
      return new byte[1]; // a zero
    // get top word;
    int top = data[len - 1];
    int add;
    if (top < 0)
      add = 5;
    else if (top > 0x7fffff)
      add = 4;
    else if (top > 0x7fff)
      add = 3;
    else if (top > 0x7f)
      add = 2;
    else
      add = 1;

    int j = (len - 1) * 4 + add;
    byte r[] = new byte[j];
    for (int i = 0; i < len - 1; ++i) {
      int d = data[i];
      r[--j] = (byte) d;
      d >>>= 8;
      r[--j] = (byte) d;
      d >>>= 8;
      r[--j] = (byte) d;
      d >>>= 8;
      r[--j] = (byte) d;
    }
    int d = data[len - 1];
    while (d != 0) {
      r[--j] = (byte) d;
      d >>>= 8;
    }

    return r;
  }

  /**
   * Internal function to parse a hex string into a byte array.
   * @param s input string, containgin a hex number
   * @return a new allocated byte array containing the number.
   */
  final private static byte[] toBytesfromHex(String s)
  {
    byte[] sb = s.getBytes();
    byte[] r = new byte[1 + (sb.length + 1) / 2];
    for (int i = 1, j = 0; j < sb.length;) {
      int b = 0;
      if (sb[j] >= '0' && sb[j] <= '9')
        b = sb[j] - '0';
      else if (sb[j] >= 'A' && sb[j] <= 'F')
        b = sb[j] - 'A' + 10;
      else if (sb[j] >= 'a' && sb[j] <= 'f')
        b = sb[j] - 'a' + 10;

      ++j;

      int c = 0;
      if (j < sb.length) {
        if (sb[j] >= '0' && sb[j] <= '9')
          c = sb[j] - '0';
        else if (sb[j] >= 'A' && sb[j] <= 'F')
          c = sb[j] - 'A' + 10;
        else if (sb[j] >= 'a' && sb[j] <= 'f')
          c = sb[j] - 'a' + 10;
      } else {
        c = b;
        b = 0;
      }
      ++j;

      r[i++] = (byte) ((b << 4) | c);
    }
    return r;
  }

  /**
   * Internal function to parse a decimal string into a byte array.
   * @param s input string, containgin a decimal number
   * @return a new allocated byte array containing the number.
   */
  final private static byte[] toBytes(String s)
  {
    IntImpl r = ZERO;
    for (int i = 0; i < s.length(); ++i)
      r = r.multiply(TEN).add(new IntImpl((long) (s.charAt(i) - '0')));
    return r.toByteArray();
  }

  /**
   * Subtract from this number the value o, and return a new BigInteger containing the result.
   * pseudo: return this - o
   * @param o the subtracted value
   */
  public IntImpl subtract(IntImpl o)
  {
    if (positive != o.positive) {
      IntImpl t = this;
      boolean sign = positive;
      if (positive)
        o = o.negate();
      else
        t = this.negate();
      t = t.add(o);
      t.positive = sign;
      return t;
    }
    if (len < o.len)
      return o.subtract(this).negate();

    int[] res = new int[len];
    int cc = sub(res, data, len, o.data, o.len);
    if (cc != len) {
      IntImpl r = new IntImpl(res, len, !positive);
      r.fixFF();
      return r;
    }
    return new IntImpl(res, len, positive);
  }

  /**
   * Internal function to perform a subtraction on big integer arrays.   
   * Pseudeo: res = a - b
   * @param res int array for the result. This array must be large enough to keep the result (al+1).
   * @param a int array of first operand.
   * @param al length of first operand. Regard: al >= bl.
   * @param b int array of second operand.
   * @param bl length of second operand.
   */
  private static int sub(int[] res, int[] a, int al, int b[], int bl)
  {
    long carry = 0;
    int i;
    for (i = 0; i < bl; ++i) {
      carry += (a[i] & 0xffffffffL) - (b[i] & 0xffffffffL);
      res[i] = (int) carry;
      carry >>= 32;
    }
    for (; carry != 0 && i < al; ++i) {
      carry += a[i] & 0xffffffffL;
      res[i] = (int) carry;
      carry >>= 32;
    }

    // only executed if carry is false    
    if (res != a)
      for (; i < al; ++i)
        res[i] = a[i];

    al += (int) carry;
    return al;
  }

  /**
   * Add to this number the value o, and return a new BigInteger containing the result.
   * pseudo: return this + o
   * @param o the added value
   */
  IntImpl add(IntImpl o)
  {
    if (positive != o.positive) {
      IntImpl r = (IntImpl) o.clone();
      r.positive = !r.positive;
      return subtract(r);
    }

    if (len < o.len)
      return o.add(this);

    int[] res = new int[len + 1];
    int nl = add(res, data, len, o.data, o.len);
    return new IntImpl(res, nl, positive);
  }

  /**
   * Internal function to perform an addition on big integer arrays. 
   * Pseudeo: res = a + b
   * @param res int array for the result. This array must be large enough to keep the result (al+1).
   * @param a int array of first operand. 
   * @param al length of first operand. Regard: al >= bl.
   * @param b int array of second operand.
   * @param bl length of second operand.
   */
  private static int add(int[] res, int[] a, int al, int[] b, int bl)
  {
    long carry = 0;
    int i;
    for (i = 0; i < bl; ++i) {
      carry += a[i] & 0xffffffffL;
      carry += b[i] & 0xffffffffL;
      res[i] = (int) carry;
      carry >>= 32;
    }
    for (; carry != 0 && i < al; ++i) {
      carry += a[i] & 0xffffffffL;
      res[i] = (int) carry;
      carry >>= 32;
    }
    if (res != a)
      for (; i < al; ++i)
        res[i] = a[i];

    if (carry != 0)
      res[al++] = 1;

    return al;
  }

  /**
   * Multiply this number by the value o, and return a new BigInteger containing the result.
   * pseudo: return this * o
   * @param o the multiplicator.
   */
  public IntImpl multiply(IntImpl o)
  {
    int res[] = new int[len + o.len + 3];
    
    if (this == o)
      square(res, data, len);
    else
      mul(res, data, 0, len, o.data, 0, o.len);

    return new IntImpl(res, res.length, positive == o.positive);
  }

  /**
   * Internal function to perform a multiplication on big integer arrays.   
   * Pseudeo: res = a * b
   * @param res int array for the result. This array must be large enough to keep the result (al+bl).
   * @param a int array of first operand.
   * @param a0 start offset of first operand.
   * @param al length of first operand.
   * @param b int array of second operand.
   * @param b0 start offset of second operand.
   * @param bl length of second operand.
   */
  private final static int mul(int res[], int a[], int a0, int al, int b[],
      int b0, int bl)
  {
    // expanded first loop to remove the need for a clear
    long carry = 0;
    long temp = a[a0++] & 0xffffffffL;
    int i = 0;
    for (int bi = b0; i < bl; ++bi, ++i) {
      carry += (b[bi] & 0xffffffffL) * temp;
      res[i] = (int) carry;
      carry >>>= 32;
    }
    res[i] = (int) carry;

    for (int q = 1; a0 < al; ++q) {
      carry = 0;
      temp = a[a0++] & 0xffffffffL;
      i = q;
      for (int bi = b0; bi < bl; ++bi, ++i) {
        carry += (b[bi] & 0xffffffffL) * temp;
        carry += res[i] & 0xffffffffL;

        res[i] = (int) carry;
        carry >>>= 32;
      }
      res[i] = (int) carry;
    }
    int r = al + bl - 1;
    while (r >= 0 && res[r] == 0)
      --r;
    return r + 1;
  }

  /**
   * Internal function to perform a square on big integer arrays.   
   * Pseudeo: res = a * a
   * @param res int array for the result. This array must be large enough to keep the result (al+al).
   * @param a int array of first operand.
   * @param al length of first operand.
   */
  private final static void square(int dst[], int src[], int len)
  {
    // calc the squares a1*a1, a2*a2, ... - also erases the buffer
    int i = len - 1;
    int j = i + len;
    dst[j] = 0;
    for (; i >= 0; --i) {
      long p = src[i] & 0xffffffffL;
      p *= p;
      dst[j--] = (int) (p >>> 32);
      dst[j--] = (int) p;
    }

    // add the mixed terms
    for (i = 0; i < len; ++i) {
      long l = src[i] & 0xffffffffL;
      long carry = 0;
      for (j = i + 1; j < len; ++j) {
        long m = l * (src[j] & 0xffffffffL);
        carry += m + m + (dst[i + j] & 0xffffffffL);
        dst[i + j] = (int) carry;
        //if (carry + 0x8000000000000000L < m + 0x8000000000000000L)
        if (m < 0) // overflow is only caused by m!
          carry = 0x100000000L | (carry >>> 32);
        else
          carry = (carry >>> 32);
      }
      for (; carry != 0; ++j) {
        carry += (dst[i + j] & 0xffffffffL);
        dst[i + j] = (int) carry;
        carry >>>= 32;
      }
    }  
  }
  private final static void square2(int res[], int a[], int al)
  {
    // calc the squares, shifted right by one to avoid overflows
    // btw: erase the buffer
    int i = al - 1;
    int j = i + al;
    int f = 0;
    res[j] = 0;
    for (; i >= 0; --i) {
      long p = a[i] & 0xffffffffL;
      p *= p;
      res[j--] = (int) (p >>> 33) | f;
      res[j--] = (int) (p >>> 1);
      f = (int) p << 31;
    }
    // add the mixed terms
    for (i = 0; i < al; ++i) {
      long l = a[i] & 0xffffffffL;
      long carry = 0;
      for (j = i + 1; j < al; ++j) {
        long m = l * (a[j] & 0xffffffffL);
        carry += (res[i + j] & 0xffffffffL) + m;
        res[i + j] = (int) carry;
        carry >>>= 32;
      }
      for (; carry != 0; ++j) {
        carry += (res[i + j] & 0xffffffffL);
        res[i + j] = (int) carry;
        carry >>>= 32;
      }
    }
    // shift left by one and set lowest bit
    f >>>= 31;
    for (i = 0; i < al + al; ++i) {
      int t = res[i];
      res[i] = t + t + f;
      f = t >>> 31;
    }
  }

  /**
   * Divide this number by the value o, and return a new BigInteger containing the result.
   * pseudo: return this / o
   * @param div the divisor.
   */
  IntImpl divide(IntImpl div)
  {
    // handle trivial cases
    if (div.len == 0) // divide by 0
      throw new RuntimeException("division by ZERO");
    if (len == 0) // divide ZERO
      return ZERO;
    if (len < div.len) // div > this
      return ZERO;

    // calc shift
    int bits = 31 - (div.bitLength() & 31);
    if (bits < 0)
      bits += 32;

    IntImpl divShifted = div.shiftLeft(bits);
    IntImpl r = shiftLeft(bits);
    // remainder
    int[] dndVal = new int[r.len + 1];
    System.arraycopy(r.data, 0, dndVal, 0, r.len);
    // result
    int[] quotVal = new int[r.len];

    modDiv(dndVal, r.len, divShifted.data, divShifted.len, quotVal);

    return new IntImpl(quotVal, quotVal.length, positive == div.positive);
  }

  /**
   * Calculate the modulo for this number by the value o, and return a new BigInteger containing the result.
   * pseudo: return this mod o
   * @param div the modulo.
   */
  public IntImpl mod(IntImpl div)
  {
    // handle trivial cases
    if (div.len == 0) // divide by 0
      throw new RuntimeException("division by ZERO");
    if (len == 0) // divide ZERO
      return ZERO;
    if (len < div.len) // div > this
      return this;

    // calc shift
    int bits = 31 - (div.bitLength() & 31);
    if (bits < 0)
      bits += 32;

    IntImpl divShifted = div.shiftLeft(bits);
    IntImpl r = shiftLeft(bits);
    // remainder
    int[] dndVal = new int[r.len + 1];
    System.arraycopy(r.data, 0, dndVal, 0, r.len);
    // result
    int[] quotVal = new int[r.len];

    modDiv(dndVal, r.len, divShifted.data, divShifted.len, quotVal);

    return r = new IntImpl(dndVal, dndVal.length, positive == div.positive)
        .shiftRight(bits);
  }

  /**
   * Calculate the division and modulo for this number by the value o, 
   * and return a new BigInteger array containing the results.
   * pseudo: return {this / o, this mod o }
   * @param div the divisor/modulo.
   */
  IntImpl[] divideAndRemainder(IntImpl div)
  {
    // handle trivial cases
    if (div.len == 0) // divide by 0
      throw new RuntimeException("division by ZERO");
    if (len == 0) // divide ZERO
      return new IntImpl[] { ZERO, ZERO };
    if (len < div.len) // div > this
      return new IntImpl[] { ZERO, this };

    // calc shift
    int bits = 31 - (div.bitLength() & 31);
    if (bits < 0)
      bits += 32;

    IntImpl divShifted = div.shiftLeft(bits);
    IntImpl r = shiftLeft(bits);
    // remainder
    int[] dndVal = new int[r.len + 1];
    System.arraycopy(r.data, 0, dndVal, 0, r.len);
    // result
    int[] quotVal = new int[r.len];

    modDiv(dndVal, r.len, divShifted.data, divShifted.len, quotVal);

    return new IntImpl[] {
        new IntImpl(quotVal, quotVal.length, positive == div.positive),
        new IntImpl(dndVal, dndVal.length, positive == div.positive)
            .shiftRight(bits) };
  }

  /**
   * Internal function to perform a division / modulo on big integer arrays.   
   * Pseudeo: c = a / b, a = a % b
   * @param c int array for the result. This array must be large enough to keep the result (bl).
   * @param a int array of first operand.
   * @param al length of first operand.
   * @param b int array of second operand.
   * @param bl length of second operand.
   */
  private final static void modDiv(int[] a, int al, int[] b, int bl, int[] c)
  {
    int divlen = bl - 1;

    // modMSW aligned that highest bit is zero and 2nd highest bit is set
    long modMSW = b[divlen];
    long modMSW32 = modMSW << 32;
    long modLSW = divlen > 0 ? b[divlen - 1] & 0xffffffffL : 0;
    for (int i = al - bl; i >= 0;) {
      long a0 = a[al] & 0xffffffffL;
      long a01 = (a0<<32) | (a[--al] & 0xffffffffL);

      // calculate:  c0 = a0|a1 / b0
//      long c0 = a0 == modMSW ? 0xffffffffL : (a01 / modMSW) & 0xffffffffL;
      long c0;
      if (a0 == modMSW) {
        c0 = 0xffffffffL; // wirklich noetig?
      } else
      if (a01 >= 0) {
        c0 = a01 / modMSW;        
      } else {
        c0 = (((a01>>>1) / modMSW) << 1) + 1 ;
      }

      if (c0 > 0) {
        if (al > 0 && modLSW != 0) {
          // a0|a1 -= c0*b0 // a0 part always gets zero!
          a01 -= c0 * modMSW;
          // add a2 => a0|a1|a2 - c0*b0|0          
          a01 = (a01 << 32) | (a[al - 1] & 0xffffffffL);
          long c0b1 = c0 * modLSW;
          while ((a01 >= 0 && (c0b1 < 0 || (c0b1 >= 0 && a01 < c0b1)))
              || (a01 < 0 && c0b1 < 0 && a01 < c0b1)) {
            --c0;

            long t = a01 + modMSW32;
            c0b1 -= modLSW;
            if (c0b1 < 0 && a01 < 0 && t >= 0)
              break;
            a01 = t;
          }
        }

        if (c0 > 0) {
          // start of mulSub          
          int ai = al - divlen;
          long carry = 0;
          for (int bi = 0; bi <= divlen; ++bi) {
            carry += (b[bi] & 0xffffffffL) * c0;
            int t = a[ai];
            a[ai++] = t - (int) carry;
            carry = (carry >>> 32)
                + (((int) carry + 0x80000000) > (t + 0x80000000) ? 1 : 0);
          }
          if (carry != 0) {
            carry = a[ai] - carry;
            a[ai] = (int) carry;
          }
          if (carry < 0)
          //if (mulSub(al, divlen, c0, b, a)) // got negative result?
          {
            --c0;

            carry = 0;
            ai = al - divlen;
            for (int bi = 0; bi <= divlen; bi++) {
              carry += (b[bi] & 0xffffffffL);
              carry += (a[ai] & 0xffffffffL);
              a[ai++] = (int) carry;
              carry >>>= 32;
            }
            if (carry > 0)
              a[ai] = a[ai] + (int) carry;
          }
        }
      }
      c[i--] += (int) c0;
    }
  }

  /**
   * Calculate the modulo inverse for this number by the value n, and return a new BigInteger containing the result.
   * pseudo: solves a * result == 1 (mod n)
   * @param n the modulo.
   */
  IntImpl modInverse(IntImpl n)
  {
    IntImpl x = ZERO;
    IntImpl y = ONE;
    IntImpl a = this;
    IntImpl b = n;
    IntImpl d, t;
    boolean sign = true;

    while (b.len > 0) {
      d = a.divide(b);
      t = a.mod(b);

      a = b;
      b = t;

      t = d.multiply(x).add(y);

      y = x;
      x = t;
      sign = !sign;
    }
    if (!sign)
      y = n.subtract(y);

    if (a.len != 1 || a.data[0] != 1)
      return ZERO;

    return y.mod(n);
  }

  /**
   * Compare this BigInteger to another BigInteger.
   * @param o the other BigInteger.
   * @return -1 if this < o, 0 if this == o, 1 if this > o.
   */
  int compareTo(IntImpl o)
  {
    if (len > o.len)
      return 1;
    if (len < o.len)
      return -1;
    for (int i = len - 1; i >= 0; --i) {
      if (data[i] + 0x80000000 > o.data[i] + 0x80000000)
        return 1;
      if (data[i] + 0x80000000 < o.data[i] + 0x80000000)
        return -1;
    }
    return 0;
  }

  /**
   * Get the index of the lowest set bit.
   * @return The index of the lowest set bit.
   */
  int getLowestSetBit()
  {
    for (int i = 0, n = 0; i < len; ++i, n += 32) {
      int d = data[i];
      if (d != 0) {
        for (;; ++n) {
          if ((d & 1) == 1)
            return n;
          d >>= 1;
        }
      }
    }
    return -1;
  }

  /**
   * Calculate the left shift for this number by the value n, and return a new BigInteger containing the result.
   * pseudo: return this << n
   * @param n Count of left shifts.
   * @return a new allocated BigInteger containing the result.
   */
  IntImpl shiftLeft(int n)
  {
    if (n <= 0) {
      if (n == 0)
        return this;
      return shiftRight(-n);
    }

    int d[] = new int[len + ((n + 31) >>> 5)];

    long k = 0;
    int j = n >>> 5;
    n &= 31;
    for (int i = 0; i < len; ++i) {
      k >>>= 32;
      k |= (data[i] & 0xffffffffL) << n;
      d[j++] = (int) k;
    }
    if ((k >>= 32) != 0)
      d[j] = (int) k;
    return new IntImpl(d, d.length, positive);
  }

  /**
   * Calculate the rights shift for this number by the value n, and return a new BigInteger containing the result.
   * pseudo: return this >> n
   * @param n Count of right shifts.
   * @return a new allocated BigInteger containing the result.
   */
  IntImpl shiftRight(int n)
  {
    if (n <= 0) {
      if (n == 0)
        return this;
      return shiftLeft(-n);
    }
    int bytes = n >> 5;
    if (bytes >= len)
      return ZERO;

    int d[] = new int[len - bytes + 1];
    n = 32 - (n & 31);

    int i, j = bytes;
    long k = (data[j++] & 0xffffffffL) << n;
    for (i = 0; j < len; ++i) {
      k >>>= 32;
      k |= (data[j++] & 0xffffffffL) << n;
      d[i] = (int) k;
    }
    k >>= 32;
    d[i] = (int) k;

    return new IntImpl(d, d.length, positive);
  }

  /**
   * Return the count of used bits.
   * @return Return the count of used bits.
   */
  int bitLength()
  {
    if (len == 0)
      return 0;
    int w = data[len - 1];
    int n = 0;
    for (; w != 0; w >>>= 1)
      ++n;
    return (len - 1) * 32 + n;
  }

  /**
   * Performs a check whether this number is probably prime.
   * The probability is 1 - (1/2)^numChecks.
   * @param rand the used randomgenerator to perform the checks.
   * @param numChecks count of performed checks
   * @return false if it's definitely not a prime, true either.
   */
  boolean isProbablePrime(Random rand, int numChecks)
  {
    if (len == 0 || (data[0] & 1) == 0)
      return false;

    IntImpl p = this;
    IntImpl q = p.subtract(ONE);
    int nb = q.getLowestSetBit();
    IntImpl m = q.shiftRight(nb);
    IntImpl b = ONE, z;

    // first pass always runs with a 2
    int bl = 2;
    int i;
    for (i = 0; i < numChecks; ++i) {
      b = new IntImpl(bl, rand);
      bl = bitLength() - 1;

      int j = 0;
      // z darf weder 1 noch q (== p-1) sein
      z = b.modPow(m, p);
      if (z.compareTo(ONE) != 0) {
        while (z.compareTo(q) != 0) {
          // genuegend loops
          if (++j >= nb)
            break;
          // z == 1 -> erfolg
          if (z.compareTo(ONE) == 0) {
            j = nb;
            break;
          }
          z = z.multiply(z).mod(p);
        }
      }
      if (j == nb)
        break;
    }
    return i == numChecks;
  }

  /**
   * Calculate the greatest common divisor.
   * So (this / result) is prime to (b / result).
   * @param b the 2nd BigInteger.
   * @return a new BigInteger containing the greatest common divisor.
   */
  IntImpl gcd(IntImpl b)
  {
    IntImpl a = this;
    if (a.compareTo(b) < 0) {
      a = b;
      b = this;
    }
    while (b.len > 0) {
      IntImpl t = a.mod(b);
      a = b;
      b = t;
    }
    return a;
  }

  /**
   * Internal function to calculate a modInverse for the lowest int.
   * @param _x lowest int of n.
   * @return A result where: x*result mod 2^32 = 1
   */
  private static int modInverse32(int _x)
  {
    int z = 0;
    long x = _x & 0xffffffffL;
    long y = 1;
    long s = 2;
    for (int i = 2; i <= 32; ++i) {
      long ss = s << 1;
      if (s < (x * y & (ss - 1)))
        y += s;
      s = ss;
//      if ((x * y & s - 1) != 1)  y /= z;
    }
    return -(int) y;
  }

  int modInv32()
  {
    return modInverse32(data[0]);
  }
  
  IntImpl montgomery(IntImpl mod, int n0)
  {
    int res[] = new int[mod.len*2 + 1];

    IntImpl dis = this.shiftLeft(mod.len << 5);
    System.arraycopy(dis.data, 0, res, 0, dis.len);
    montgomery(res, mod.data, mod.len, n0);

    return new IntImpl(res, res.length, true);
  }
  
  /**
   * Perform the montgomery transformation.
   * @param t int array for input and the result. This array must be large enough to keep the result (ml+ml+1).
   * @param mod int array with modulo.
   * @param ml length of modulo.
   * @param n0 lowest int.
   */
  private static void montgomery(int t[], int mod[], int ml, int n0)
  {
    // t = (t + (t*n' mod r) * n )
    int ml2 = ml + ml;
    long carry = 0;
    for (int i = 0; i < ml; ++i) {
      long m = (n0 * t[i]) & 0xffffffffL;
      int k = i;
      for (int j = 0; j < ml; ++j, ++k) {
        carry += (t[k] & 0xffffffffL) + m * (mod[j] & 0xffffffffL);
        t[k] = (int) carry;
        carry >>>= 32;
      }
      for (; carry != 0 && k < ml2; ++k) {
        carry += (t[k] & 0xffffffffL);
        t[k] = (int) carry;
        carry >>>= 32;
      }
    }

    // t = t / r              = shift right
    {
      int i = 0;
      for (; i < ml; ++i)
        t[i] = t[i + ml];
      for (; i < ml2; ++i)
        t[i] = 0;
    }

    boolean sub = carry != 0;
    if (!sub) {
      int i = ml - 1;
      for (; i >= 0; --i) {
        if (t[i] != mod[i])
          break;
      }
      sub = i < 0 || t[i] + 0x80000000 > mod[i] + 0x80000000;
    }
    if (sub) {
      sub(t, t, ml, mod, ml);
    }
  }

  /**
   * Calculate this^exp % mod.
   * Right now it works only for even modulus.
   * @param exp the exponent.
   * @param mod the modulus.
   * @return a new BigInteger containing the result
   */
  public IntImpl modPow(IntImpl exp, IntImpl mod)
  {
    if (!mod.positive)
      throw new RuntimeException("modulus not positive");

    if (!exp.positive)
      return ZERO;

    if (exp.len == 0)
      return ONE;

    // check whether this is negative
    boolean negative = !positive;

    IntImpl dis = this;
    if (negative)
      dis = negate();

    // reduce size
    if (dis.compareTo(mod) >= 0)
      dis = dis.mod(mod);

    IntImpl res;
    if ((mod.data[0] & 1) == 1)
      res = dis.oddModPow(exp, mod);
    else {
      throw new RuntimeException("modulus even");
      /*      
       int lb = mod.getLowestSetBit();
       BigInteger q = mod.shiftRight(lb);
       BigInteger _2j = ONE.shiftLeft(lb);
       BigInteger x1 = dis.mod(q).oddModPow(exp, q);
       BigInteger x2 = dis.mod2(lb).modPow2(exp.mod2(lb-1), lb);
       
       BigInteger q_1 = q.modInverse(_2j);      
       BigInteger y = x2.subtract(x1).multiply(q_1).mod2(lb);
       res = x1.add(q.multiply(y));
       */
    }
    if (!negative)
      return res;

    return res.modInverse(mod);
  }

  /**
   * Calculate this^exp % mod for odd modulus.
   * @param exp the exponent.
   * @param mod the modulus.
   * @return a new BigInteger containing the result
   */
  private IntImpl oddModPow(IntImpl exp, IntImpl mod)
  {
    // get the count of exponent bits
    int ipr = exp.bitLength();

    // m = 2**bits, where m*m = sqrt(ipr/2)    
    int bits = 1;
    for (int i = ipr / 2; i != 0; i >>>= 1)
      ++bits;
    bits >>= 1; 
    if (bits == 0)
      bits = 1;

    // size for array of precalculated powers of a
    int pack = (1 << bits);

    int ibit = ipr & 31;
    ipr = ipr >>> 5;

    if (ibit == 0) {
      if (ipr == 0)
        return ONE;
      ibit = 32;
      --ipr;
    }

    int bl = mod.len << 5;
    int sz = mod.len * 2; // + 1;

    int n0 = modInverse32(mod.data[0]);

    IntImpl dis = shiftLeft(bl).mod(mod);
    IntImpl res = ONE.shiftLeft(bl).mod(mod);

    //System.out.println("n: " + dis.toHexString());
    //System.out.println("1: " + res.toHexString());
    
    // calculate 1, a % mod, a*a % mod, a*a*a % mod, a^4 % mod, usw
    int[][] as = new int[pack][];
    as[0] = res.data;
//    as[0] = new int[sz];
//    System.arraycopy(res.data, 0, as[0], 0, res.data.length);
    as[1] = dis.data;
//    as[1] = new int[sz];
//    System.arraycopy(dis.data, 0, as[1], 0, dis.data.length);
    for (int i = 2; i < pack; ++i) {
      // skip even values
      // if ( (i & 1) == 0) continue;
      int[] t1 = new int[sz];
      if ( (i&1) == 1)
        mul(t1, dis.data, 0, mod.len, as[i - 1], 0, mod.len);
      else
        square(t1, as[i>>1], mod.len);
      montgomery(t1, mod.data, mod.len, n0);
      as[i] = t1;
    }

    int[] t1 = new int[sz];
    int[] t2 = new int[sz];
    System.arraycopy(res.data, 0, t1, 0, mod.len);
    // t1 = res.data;

    int u[] = exp.data; // index into exponent
    int mask = (u[ipr] << (32 - ibit));
    int ubits = 0, ui = 0;
    boolean isOne = false;
    while (res.len != 0) // quick exit on 0
    {
      ui <<= 1;
      if (mask < 0)
        ui |= 1;
      mask <<= 1;
      // delay window use un til a bit is set inside
      if (ui != 0) ++ ubits;
      if (ubits == bits) {
        //System.out.println(ui);
        //System.out.println("dat: " + new IntImpl(as[ui], as[ui].length, true).toHexString());      
        mul(t2, t1, 0, mod.len, as[ui], 0, mod.len);
        //System.out.println("mul: " + new IntImpl(t2, t2.length, true).toHexString());      
        montgomery(t2, mod.data, mod.len, n0);
        //System.out.println("mon: " + new IntImpl(t2, t2.length, true).toHexString());      
        int[] t = t2;
        t2 = t1;
        t1 = t;

        ubits = 0;
        ui = 0;
        isOne = false;
      }

      //System.out.println("r: " + new IntImpl(t1, t1.length, true).toHexString());      
      
      if (--ibit == 0) {
        if (ipr-- == 0)
          break;
        ibit = 32;
        mask = u[ipr];
      }
      if (isOne) continue;
      
      square(t2, t1, mod.len);
      montgomery(t2, mod.data, mod.len, n0);
      int[] t = t2;
      t2 = t1;
      t1 = t;
    }

    if (ubits != 0 && ui != 0) {
//      for (int ii = 0; ii < sz; ++ii) t2[ii] = 0;
      mul(t2, t1, 0, mod.len, as[ui], 0, mod.len);
      montgomery(t2, mod.data, mod.len, n0);
      int[] t = t2;
      t2 = t1;
      t1 = t;
    }

    montgomery(t1, mod.data, mod.len, n0);

    return new IntImpl(t1, mod.len, positive);
  }
/*
  static String display(int [] t1, IntImpl mod, int n0)
  {
    int tt[] = (int[]) t1.clone();
    montgomery(tt, mod.data, mod.len, n0);
    return new IntImpl(tt, mod.len, true).toString();
  }
  */
  /**
   * Create a BigInteger with - this value. x.negate() + x = 0
   * @return a new BigInteger with the negaitve value of this.
   */
  IntImpl negate()
  {
    IntImpl r = (IntImpl) clone();
    r.positive = !positive;
    return r;
  }

  /* 
   The following functions are only necessary, when this class is derived from Number 
   */
  /**
   * Return the lowest integer as an int.
   * @return the lowest integer as an int.
   * /
   int intValue()
   {
   return len > 0 ? data[0] : 0;
   }
   /**
   * Return the lowest two integers as a long.
   * @return the lowest two integers as a long.
   * /
   long longValue()
   {
   long l = 0;
   if (len > 0)
   l = data[0]&0xffffffffL;
   if (len > 1)
   l |= (data[0]&0xffffffffL)<<32;
   return l;
   }
   /**
   * Return the lowest two integers as a float.
   * @return the lowest two integers as a float.
   * /
   float floatValue() {
   return (float)longValue();
   }
   /**
   * Return the lowest two integers as a double.
   * @return the lowest two integers as a double.
   * /
   double doubleValue() {
   return (double)longValue();
   }
   /**
   * Checks whether the other object is an BigInteger. If true the content is compared to this value.
   * @param x the other object.
   * @return true if the other object is an BigInteger and contains the same value as this.
   * /
   boolean equals(Object x)
   {
   if (x instanceof BigInteger)
   return compareTo((BigInteger)x)==0;
   return false;
   }
   
   /*
   The following functions are not needed for cryptography. They can be omitted, and not yet fully tested and commented.
   * /
   BigInteger abs()
   {
   BigInteger r = (BigInteger)clone();
   r.positive = true;
   return r;
   }
   
   BigInteger and(BigInteger o)
   {
   int l = len < o.len ? len : o.len;
   int r[] = new int[l];
   for (int i = 0; i < l; ++i)
   {
   r[i] = data[i] & o.data[i];
   }
   return new BigInteger(r, l, true);
   }
   BigInteger andNot(BigInteger o)
   {
   int l = len < o.len ? len : o.len;
   int m = len > o.len ? len : o.len;
   int d[] = len > o.len ? data : o.data;
   int r[] = new int[l];
   int i = 0;
   for (; i < l; ++i)
   {
   r[i] = data[i] & ~o.data[i];
   }
   for (; i < m; ++i)
   r[i] = d[i];
   return new BigInteger(r, m, true);
   }
   int bitCount()
   {
   int r = 0;
   for (int i = 0; i < len; ++i)
   {
   int a = data[i]&0x55555555;
   int b = (data[i]&0xaaaaaaaa)>>>1;
   a += b;
   
   b = (a&0xcccccccc)>>2;
   a &= 0x33333333;
   a += b;
   
   b = (a&0xf0f0f0f0)>>4;
   a &= 0x0f0f0f0f;
   a += b;
   
   b = (a&0xff00ff00)>>8;
   a &= 0x00ff00ff;
   a += b;
   
   b = (a&0xffff0000)>>16;
   a &= 0x0000ffff;
   a += b;
   
   r += a;
   }
   return r;
   }
   BigInteger clearBit(int n) throws ArithmeticException
   {
   if (n < 0) throw new ArithmeticException("n < 0");
   int o = n>>>5;
   n &= 31;
   int res[] = new int[len];
   System.arraycopy(data, 0, res, 0, len);
   res[o] &= ~(1<<n);
   return new BigInteger(res, len, positive);
   }
   BigInteger flipBit(int n) throws ArithmeticException
   {
   if (n < 0) throw new ArithmeticException("n < 0");
   int o = n>>>5;
   n &= 31;
   int res[] = new int[len];
   System.arraycopy(data, 0, res, 0, len);
   res[o] ^= (1<<n);
   return new BigInteger(res, len, positive);
   }    
   int hashCode()
   {
   int r = len;
   for (int i = 0; i < len; ++i)
   r ^= data[i];
   return r;
   }
   BigInteger max(BigInteger o)
   {
   if (compareTo(o) >= 0)
   return this;
   return o;
   }
   BigInteger min(BigInteger o)
   {
   if (compareTo(o) <= 0)
   return this;
   return o;
   }
   BigInteger not()
   {
   int r[] = new int[len];
   
   for (int i = 0; i < len; ++i)
   {
   r[i] = ~data[i];
   }
   return new BigInteger(r, len, true);
   }
   BigInteger or(BigInteger o)
   {
   int l = len < o.len ? len : o.len;
   int r[] = new int[l];
   for (int i = 0; i < l; ++i)
   {
   r[i] = data[i] | o.data[i];
   }
   return new BigInteger(r, l, true);
   }
   BigInteger remainder(BigInteger val)
   {
   return mod(val);
   }
   BigInteger setBit(int n) throws ArithmeticException
   {
   if (n < 0) throw new ArithmeticException("n < 0");
   int o = n>>>5;
   n &= 31;
   int res[] = new int[len];
   System.arraycopy(data, 0, res, 0, len);
   res[o] |= (1<<n);
   return new BigInteger(res, len, positive);
   }    
   int signum() {
   if (len == 0)
   return 0;
   return positive ? 1 : -1;
   }
   boolean testBit(int n) throws ArithmeticException
   {
   if (n < 0) throw new ArithmeticException("n < 0");
   int o = n>>>5;
   n &= 31;
   return (data[o] & (1<<n)) != 0;
   }    
   String toString(int radix)
   {
   if (radix == 16)
   return toHexString();
   return toString();
   }
   static BigInteger valueOf(long val)
   {
   return new BigInteger(val);
   }
   BigInteger xor(BigInteger o)
   {
   int l = len < o.len ? len : o.len;
   int r[] = new int[l];
   for (int i = 0; i < l; ++i)
   {
   r[i] = data[i] ^ o.data[i];
   }
   return new BigInteger(r, l, true);
   }
   BigInteger pow(int exponent) throws ArithmeticException
   {
   if (exponent < 0)
   throw new ArithmeticException();
   if (exponent == 0)
   return ONE;
   BigInteger res = ONE;
   int loop = 32;
   while (exponent > 0)
   {
   --loop;
   exponent <<= 1;
   }
   while (loop-- > 0)
   {
   if (exponent < 0)
   res = res.multiply(this);
   if (loop > 0)
   res = res.multiply(res);
   exponent <<= 1;
   }
   return res;
   }
   /**/

}

/*
 * $Log: BigInteger.java,v $
 * Revision 1.4  2003/01/04 12:22:35  bebbo
 * @R removed DEBUG messages
 *
 * Revision 1.3  2003/01/04 12:10:37  bebbo
 * @I cleaned up imports and formatted source code
 *
 * Revision 1.2  2003/01/04 12:07:15  bebbo
 * @R cleanup
 *
 * Revision 1.1  2000/11/15 09:14:04  bebbo
 * @N extra versions for miniSSL
 *
 * Revision 1.1  2000/11/13 22:15:58  bebbo
 * @N added to midissl
 *
 * Revision 1.5  2000/11/08 17:57:18  bebbo
 * @R fixed comments and removed korma
 *
 * Revision 1.4  2000/11/08 17:55:14  bebbo
 * @R disabled korma
 *
 * Revision 1.3  2000/11/03 17:37:19  bebbo
 * @N added lot of comments
 *
 * Revision 1.2  2000/11/03 10:22:56  bebbo
 * @N added montgomery mulmod
 *
 * Revision 1.1  2000/10/17 22:08:41  bebbo
 * @N first working version
 *
 */